#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>


/* Partition table parsing code taken from Rockbox */

#define SECTOR_SIZE 512

struct partinfo {
  unsigned long start; /* first sector (LBA) */
  unsigned long size;  /* number of sectors */
  unsigned char type;
};

#define BYTES2INT32(array,pos)					\
    ((long)array[pos] | ((long)array[pos+1] << 8 ) |		\
     ((long)array[pos+2] << 16 ) | ((long)array[pos+3] << 24 ))

void display_partinfo(struct partinfo* pinfo)
{
    int i;

    printf("Part    Type   Start Sector    End Sector    Size (MB)\n");
    for ( i=0; i<4; i++ ) {
        if (pinfo[i].start != 0) {
            printf("   %d    0x%02x     %10ld    %10ld  %10ld\n",i,pinfo[i].type,pinfo[i].start,pinfo[i].start+pinfo[i].size-1,pinfo[i].size/2048);
        }
    }
}


int read_partinfo(int fd, struct partinfo* pinfo)
{
    int i;
    unsigned char sector[SECTOR_SIZE];

    int res;

    res=lseek(fd,SEEK_SET,0);
    res=read(fd,sector,SECTOR_SIZE);

    /* check that the boot sector is initialized */
    if ( (sector[510] != 0x55) ||
         (sector[511] != 0xaa)) {
        fprintf(stderr,"Bad boot sector signature\n");
        return 0;
    }

    /* parse partitions */
    for ( i=0; i<4; i++ ) {
        unsigned char* ptr = sector + 0x1be + 16*i;
        pinfo[i].type  = ptr[4];
        pinfo[i].start = BYTES2INT32(ptr, 8);
        pinfo[i].size  = BYTES2INT32(ptr, 12);

        /* extended? */
        if ( pinfo[i].type == 5 ) {
            /* not handled yet */
        }
    }

    return 0;
}

int disk_read(int fd, int outfile,unsigned long start, unsigned long count)
{
    int res;
    int i;
    unsigned char buf[SECTOR_SIZE];

    fprintf(stderr,"[INFO] Seeking to sector %ld\n",start);

    res=lseek(fd,SEEK_SET,SECTOR_SIZE*start);

    if (res < 0) {
        fprintf(stderr,"[WARN] Seek error - reading %ld sectors instead\n",start);

        for (i=0;i<start;i++) {
            res=read(fd,buf,SECTOR_SIZE);

            if (res < 0) {
                perror("[ERR]  read in disk_read");
                return -1;
            }
        }
    }

    fprintf(stderr,"[INFO] Writing %ld sectors to output file\n",count);
    for (i=0;i<count;i++) {
        res=read(fd,buf,SECTOR_SIZE);

        if (res < 0) {
            perror("[ERR]  read in disk_read");
            return -1;
        }

        if (res != SECTOR_SIZE) {
            fprintf(stderr,"Short read - requested %d, received %d - aborting.\n",SECTOR_SIZE,res);
            return -1;
        }

        res=write(outfile,buf,SECTOR_SIZE);

        if (res < 0) {
            perror("[ERR]  write in disk_read");
            return -1;
        }

        if (res != SECTOR_SIZE) {
            fprintf(stderr,"Short write - requested %d, received %d - aborting.\n",SECTOR_SIZE,res);
            return -1;
        }
    }

    fprintf(stderr,"[INFO] Done.\n");
    return 0;
}

int disk_write(int fd, int infile,unsigned long start)
{
    int res;
    int i;
    int n;
    int bytesread;
    int eof;
    unsigned char buf[SECTOR_SIZE];

    fprintf(stderr,"[INFO] Seeking to sector %ld\n",start);

    res=lseek(fd,SEEK_SET,SECTOR_SIZE*start);

    if (res < 0) {
        fprintf(stderr,"[WARN] Seek error - reading %ld sectors instead\n",start);

        for (i=0;i<start;i++) {
            res=read(fd,buf,SECTOR_SIZE);

            if (res < 0) {
                perror("[ERR]  read in disk_write");
                return -1;
            }
        }
    }

    fprintf(stderr,"[INFO] Writing input file to device\n");
    bytesread=0;
    eof=0;
    while (!eof) {
        n=read(infile,buf,SECTOR_SIZE);

        if (n < 0) {
            perror("[ERR]  read in disk_write");
            return -1;
        }

        if (n < SECTOR_SIZE) eof=1;

        bytesread+=n;
        res=write(fd,buf,n);

        if (res < 0) {
            perror("[ERR]  write in disk_write");
            return -1;
        }

        if (res != n) {
            fprintf(stderr,"Short write - requested %d, received %d - aborting.\n",SECTOR_SIZE,res);
            return -1;
        }
    }

    fprintf(stderr,"[INFO] Wrote %d bytes.\n",bytesread);
    return 0;
}


void print_usage(void) {
    fprintf(stderr,"Usage: diskdump  [-i|r|w|t] [-p n] /dev/device [file]\n");
    fprintf(stderr,"         -i    Display partition information for device\n");
    fprintf(stderr,"         -r    Read from device to file\n");
    fprintf(stderr,"         -w    Write file to device\n");
    fprintf(stderr,"         -t    Test write, do not write to partition\n");
    fprintf(stderr,"         -p n  Use partition n [default: 0]\n");
    fprintf(stderr,"\n");
}

enum {
   NONE,
   SHOW_INFO,
   READ,
   WRITE,
   TEST
};

int main(int argc, char* argv[])
{
    int i;
    struct partinfo pinfo[4]; /* space for 4 partitions on 1 drive */
    int res;
    int fd;
    int outfile;
    int infile;
    int mode=SHOW_INFO;
    int p=0;
    char* device=NULL;
    char* filename=NULL;

    fprintf(stderr,"diskdump v0.1 - (C) Dave Chapman 2005\n");
    fprintf(stderr,"This is free software; see the source for copying conditions.  There is NO\n");
    fprintf(stderr,"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n");
    
    if (argc<=1) {
        print_usage();
        return 1;
    }

    i=1;
    while (i < argc) {
        if (strncmp(argv[i],"-i",2)==0) {
            mode=SHOW_INFO;
        } else if (strncmp(argv[i],"-r",2)==0) {
            mode=READ;
        } else if (strncmp(argv[i],"-w",2)==0) {
            mode=WRITE;
        } else if (strncmp(argv[i],"-t",2)==0) {
            mode=TEST;
        } else if (strncmp(argv[i],"-p",2)==0) {
            i++;
            if (i < argc) {
                p=atoi(argv[i]);
                if ((p < 0) || (p > 3)) {
                    fprintf(stderr,"Invalid partition number - must be in range 0 to 3\n");
                    return 1;
                }
            } else {
                fprintf(stderr,"Ran out of arguments: -p\n");
                return 1;
            }
        } else {
            if (argv[i][0]=='-') {
                fprintf(stderr,"Unknown option %s\n",argv[i]);
                return 1;
            } else {
                if (device==NULL) {
                   device=argv[i];
                } else if (filename==NULL) {
                   filename=argv[i];
                } else {
                   fprintf(stderr,"Too many arguments: %s\n",argv[i]);
                   return 1;
                }
            }
        }
        i++;
    }

    if ((mode==NONE) || (device==NULL) || ((mode!=SHOW_INFO) && (filename==NULL))) {
        print_usage();
        return 1;
    }

    fprintf(stderr,"[INFO] Reading partition table...\n");
    fd=open(device,O_RDONLY);
    if (fd < 0) {
        perror(device);
        return 2;
    }
    read_partinfo(fd,pinfo);
    close(fd);

    display_partinfo(pinfo);

    if (pinfo[p].start==0) {
        fprintf(stderr,"[ERR]  Specified partition (%d) does not exist:\n",p);
        display_partinfo(pinfo);
        return 3;
    }

    if (mode==READ) {
        fd=open(device,O_RDONLY);
        if (fd < 0) {
            perror(device);
            return 2;
        }

        outfile=open(filename,O_CREAT|O_WRONLY,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
        if (outfile < 0) {
           perror(filename);
           return 4;
        }

        res=disk_read(fd,outfile,pinfo[p].start,pinfo[p].size);

        close(outfile);
        close(fd);
    } else if (mode==WRITE) {
        fd=open(device,O_RDWR);
        if (fd < 0) {
            perror(device);
            return 2;
        }

        infile=open(filename,O_RDONLY);
        if (infile < 0) {
           perror(filename);
           return 4;
        }

        /* TODO: Check filesize is <= partition size */

        res=disk_write(fd,infile,pinfo[p].start);

        close(infile);
        close(fd);
    }
    return 0;
}
